import metallurgy as mg


def test_alloy_creation():

    alloy_from_string = mg.Alloy("Fe20Ag80")

    alloy_from_dict = mg.Alloy({"Fe": 20, "Ag": 80})

    assert alloy_from_string == alloy_from_dict


def test_alloy_to_string():
    alloy = mg.Alloy({"Fe": 20, "Ag": 80})
    assert alloy.to_string() == "Ag80Fe20"


def test_alloy_element_deletion():

    num_alloys = 100
    random_alloys = mg.generate.random_alloys(
        num_alloys,
        min_elements=2,
        max_elements=2,
    )

    for alloy in random_alloys:
        alloy.remove_element(alloy.elements[0])
        assert alloy.to_string() == alloy.elements[0] + "100"

    random_alloys = mg.generate.random_alloys(
        num_alloys,
        min_elements=2,
        max_elements=2,
    )

    for alloy in random_alloys:
        del alloy.composition[alloy.elements[0]]
        assert alloy.to_string() == alloy.elements[0] + "100"


def test_alloy_equality():

    assert mg.Alloy("Fe20Ag80") == "Fe20Ag80" == mg.Alloy("Ag80Fe20")


def test_sub_alloy_parsing():

    assert (
        mg.Alloy("(Fe20Ag80)50Xe50")
        == mg.Alloy("Xe50(Fe20Ag80)50")
        == "Fe10Ag40Xe50"
    )


def test_multiple_round():

    assert mg.alloy.multiple_round(0.015, 0.02) == 0.02
    assert mg.alloy.multiple_round(0.015, 0.01) == 0.02
    assert mg.alloy.multiple_round(0.12345, 0.03) == 0.12
    assert mg.alloy.multiple_round(0.12345, 0.2) == 0.2


def test_structure():

    assert mg.Alloy("Cu", "A1") == "Cu100[A1]"
    assert mg.Alloy("CuZr", "A1") == "Cu50Zr50[A1]"
    assert mg.Alloy("CuZr", "B2") == "Cu50Zr50[B2]"
    assert mg.Alloy("CuZr", "L1_0") == "Cu50Zr50[L1_0]"
    assert mg.Alloy("ZrCu", "D2_1") == "Cu85.72Zr14.28[D2_1]"
    assert mg.Alloy("CuZr", "D2_1") == "Zr85.72Cu14.28[D2_1]"
    assert (
        mg.Alloy("Fe38As38Cu12Sc12[D0_3]").to_string()
        == "Fe38As38Cu12Sc12[D0_3]"
    )

    assert (
        mg.Alloy(
            {"Fe": 0.6666666666666666, "O": 0.3333333333333333},
            structure="BaPtSb",
        )
        == "Fe66.667O33.333[BaPtSb]"
    )

    assert mg.Alloy("CuZr", "D5_3") == "Cu70Zr30[D5_3]"


def test_structure_change():

    a = mg.Alloy("CuZr", "A1")
    a.structure = "A2"
    assert a == "Cu50Zr50[A2]"

    a = mg.Alloy("CuZr", "B1")
    a.structure = "D0_23"
    assert a == "Cu75Zr25[D0_23]"


def test_structure_element_order():
    assert mg.Alloy("Cu", "A1").elements_structure_order == ["Cu"]
    assert mg.Alloy("CuZr", "B1").elements_structure_order == ["Cu", "Zr"]
    assert mg.Alloy("CuZr", "C32").elements_structure_order == ["Cu", "Zr"]
    assert mg.Alloy("Cu66.67Mn33.33", "C32").elements_structure_order == [
        "Mn",
        "Cu",
    ]
    assert mg.Alloy("CuZr", "D5_3").elements_structure_order == ["Cu", "Zr"]


def test_scale_conversion():

    assert mg.Alloy("Cu50Zr50") == mg.Alloy("Cu0.5Zr0.5")
    assert mg.Alloy("Cu99.9Zr0.1") == mg.Alloy("Cu0.999Zr0.001")



"""Utility functions handling property units and property name formatting."""

# A variety of units for some alloy properties
pretty_names = {
    "Dmax": "$D_{\mathrm{max}}$",
    "Tl": "$T_l$",
    "Tg": "$T_g$",
    "Tx": "$T_x$",
    "deltaT": "$\Delta T$",
}

units = {
    "Dmax": "mm",
    "Tl": "K",
    "Tg": "K",
    "Tx": "K",
    "deltaT": "K",
    "price_linearmix": "\\$/kg",
    "price": "\\$/kg",
    "mixing_enthalpy": "kJ/mol",
    "mixing_Gibbs_free_energy": "kJ/mol",
}
inverse_units = {}
for property_name in units:
    if "/" not in units[property_name]:
        inverse_units[property_name] = "1/" + units[property_name]
    else:
        split_units = units[property_name].split("/")
        inverse_units[property_name] = split_units[1] + "/" + split_units[0]


def pretty_name(property_name: str) -> str:
    """Returns a nicely formatted version of an alloy property name

    Parameters
    ==========

    property_name : str
        Name of the property to format.

    """
    name = ""
    property_name_parts = property_name.split("_")
    if "linearmix" in property_name or "deviation" in property_name:
        if len(property_name_parts) > 1:
            if property_name_parts[-1] == "linearmix":
                name = r"$\Sigma$ "
            elif property_name_parts[-1] == "deviation":
                name = r"$\delta$ "
            name += " ".join(
                word.title() for word in property_name_parts[0:-1]
            )
    elif property_name in pretty_names:
        name += pretty_names[property_name]
    else:
        name += " ".join(word.title() for word in property_name_parts)
    return name



from typing import Union, List
from numbers import Number
from collections.abc import Iterable

import metallurgy as mg


def shell_valence_electron_concentration_ratio(
    alloy: Union[mg.Alloy, str, dict],
    period: Union[Number, None] = None,
    valence_electrons: Union[Number, None] = None,
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the ratio between shell number and valence electron concentration
    of an alloy.

    :group: calculations.ratios

    Parameters
    ----------

    alloy : mg.Alloy, str, dict
        The alloy for which to calculate the ratio.
    period : Number
        The effective period of the alloy.
    valence_electrons : Number
        The effective number of valence electrons of the alloy.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [shell_valence_electron_concentration_ratio(a) for a in alloy]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    if period is None:
        period = mg.linear_mixture(alloy, "period")
    if valence_electrons is None:
        valence_electrons = mg.linear_mixture(alloy, "valence_electrons")

    return period / valence_electrons


def shell_mendeleev_number_ratio(
    alloy: Union[mg.Alloy, str, dict],
    period: Union[Number, None] = None,
    mendeleev_number: Union[Number, None] = None,
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the ratio between shell number and Mendeleev number of an alloy.

    :group: calculations.ratios

    Parameters
    ----------

    alloy : mg.Alloy, str, dict
        The alloy for which to calculate the ratio.
    period : Number
        The effective period of the alloy.
    mendeleev_number : Number
        The effective Mendeleev number of the alloy.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [shell_mendeleev_number_ratio(a) for a in alloy]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    if period is None:
        period = mg.linear_mixture(alloy, "period")
    if mendeleev_number is None:
        mendeleev_number = mg.linear_mixture(
            alloy, "mendeleev_universal_sequence"
        )

    if period is None or mendeleev_number is None:
        return None

    return period / mendeleev_number



"""Valence related calculations"""

from typing import Union, List
from collections.abc import Iterable
from numbers import Number

import metallurgy as mg


def valence_number(
    alloy: Union[mg.Alloy, str, dict], orbital: str
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the number of the valence electrons which are in a particular
    orbital.

    :group: calculations.valence

    Parameters
    ----------

    alloy : mg.Alloy, str, dict
        The alloy for which to calculate the valence proportion.

    orbital : str
        The orbital to calculate the proportion of.
    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [valence_number(a, orbital) for a in alloy]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    proportion = valence_proportion(alloy, orbital)

    return proportion * mg.linear_mixture(alloy, "valence_electrons")


def valence_proportion(
    alloy: Union[mg.Alloy, str, dict], orbital: str
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the proportion of the valence electrons which are in a particular
    orbital.

    :group: calculations.valence

    Parameters
    ----------

    alloy : mg.Alloy, str, dict
        The alloy for which to calculate the valence proportion.

    orbital : str
        The orbital to calculate the proportion of.
    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [valence_proportion(a, orbital) for a in alloy]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    orbital_count = {}
    for element in alloy.elements:
        orbital_count[element] = 0

        valence_electrons = mg.periodic_table.elements[element][
            "valence_electrons"
        ]
        orbitals = mg.periodic_table.elements[element]["orbitals"]

        i = 0
        electrons = 0
        while electrons < valence_electrons:
            electrons += orbitals[-1 - i]["electrons"]
            if orbitals[-1 - i]["orbital"][-1] == orbital:
                orbital_count[element] += orbitals[-1 - i]["electrons"]
            i += 1

    total_valence = mg.linear_mixture(alloy, "valence_electrons")

    if total_valence > 0:
        total = 0
        for element in alloy.elements:
            total += orbital_count[element] * alloy.composition[element]
        return total / total_valence

    return 0


def s_valence(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the proportion of the valence electrons which are in the s
    orbital.

    :group: calculations.valence

    Parameters
    ----------

    alloy : mg.Alloy, str, dict
        The alloy for which to calculate the s-valence proportion.

    """
    return valence_proportion(alloy, "s")


def p_valence(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the proportion of the valence electrons which are in the p
    orbital.

    :group: calculations.valence

    Parameters
    ----------

    alloy : mg.Alloy, str, dict
        The alloy for which to calculate the p-valence proportion.

    """
    return valence_proportion(alloy, "p")


def d_valence(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the proportion of the valence electrons which are in the d
    orbital.

    :group: calculations.valence

    Parameters
    ----------

    alloy : mg.Alloy, str, dict
        The alloy for which to calculate the d-valence proportion.

    """
    return valence_proportion(alloy, "d")


def f_valence(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the proportion of the valence electrons which are in the f
    orbital.

    :group: calculations.valence

    Parameters
    ----------

    alloy : mg.Alloy, str, dict
        The alloy for which to calculate the f-valence proportion.

    """
    return valence_proportion(alloy, "f")


def s_valence_number(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the number of valence electrons which are in the s
    orbital.

    :group: calculations.valence

    Parameters
    ----------

    alloy : mg.Alloy, str, dict
        The alloy for which to calculate the s-valence proportion.

    """
    return valence_number(alloy, "s")


def p_valence_number(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the number of the valence electrons which are in the p
    orbital.

    :group: calculations.valence

    Parameters
    ----------

    alloy : mg.Alloy, str, dict
        The alloy for which to calculate the p-valence number.

    """
    return valence_number(alloy, "p")


def d_valence_number(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the number of the valence electrons which are in the d
    orbital.

    :group: calculations.valence

    Parameters
    ----------

    alloy : mg.Alloy, str, dict
        The alloy for which to calculate the d-valence number.

    """
    return valence_number(alloy, "d")


def f_valence_number(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the number of the valence electrons which are in the f
    orbital.

    :group: calculations.valence

    Parameters
    ----------

    alloy : mg.Alloy, str, dict
        The alloy for which to calculate the f-valence number.

    """
    return valence_number(alloy, "f")



import copy
import re
from numbers import Number
from typing import List, Optional, Union

import elementy
import numpy as np

import metallurgy as mg

from .prototypes import get_random_prototype


def random_alloy(
    min_elements: int = 1,
    max_elements: int = 10,
    percentage_constraints={},
    percentage_step: float = 0.01,
    allowed_elements=[e for e in elementy.PeriodicTable().elements],
    constrain_alloy: bool = False,
    structure: bool = False,
):
    """Generate a random alloy.

    :group: alloy.generate

    Parameters
    ----------

    min_elements
        Minimum number of elements in the random alloy.

    """

    if isinstance(percentage_constraints, (list)):
        parsed_percentage_constraints = {}
        for element in percentage_constraints:
            parsed_percentage_constraints[element] = {}
        percentage_constraints = parsed_percentage_constraints

    if min_elements < max_elements:
        num_extra_elements = np.random.randint(min_elements, max_elements + 1)
    else:
        num_extra_elements = max_elements
    num_extra_elements -= len(percentage_constraints)

    if num_extra_elements > 0:
        other_elements = allowed_elements[:]
        for element in percentage_constraints:
            if element in other_elements:
                other_elements.remove(element)

        elements = list(percentage_constraints.keys()) + list(
            np.random.choice(
                other_elements,
                min(num_extra_elements, len(other_elements)),
                replace=False,
            )
        )

    elif max_elements == min_elements:
        if max_elements == len(list(percentage_constraints.keys())):
            elements = list(percentage_constraints.keys())
        elif max_elements == len(allowed_elements):
            elements = allowed_elements[:]

    else:
        elements = list(
            np.random.choice(
                list(percentage_constraints.keys()),
                num_extra_elements + len(percentage_constraints.keys()),
                replace=False,
            )
        )

    if not structure:
        percentages = np.random.rand(len(elements))
        percentages = (
            percentages
            / percentages.sum()
            * (1 - percentage_step * len(elements))
        )
        percentages += percentage_step

        composition = {}
        for j in range(len(elements)):
            composition[elements[j]] = percentages[j]
    else:
        composition = "".join(elements)

        composition += "[" + get_random_prototype().name + "]"

    alloy = mg.Alloy(
        composition,
        constraints={
            "percentages": percentage_constraints,
            "percentage_step": percentage_step,
            "min_elements": min_elements,
            "max_elements": max_elements,
            "allowed_elements": allowed_elements,
        },
    )

    if not constrain_alloy:
        alloy.constraints = None

    return alloy


def random_alloys(
    num_alloys: int,
    min_elements: int = 1,
    max_elements: int = 10,
    percentage_constraints: dict = {},
    percentage_step=0.01,
    allowed_elements: list = [e for e in elementy.PeriodicTable().elements],
    constrain_alloys=False,
    structures=False,
):
    """Generate multiple random alloys.

    :group: alloy.generate

    Parameters
    ----------

    min_elements
        Minimum number of elements in the random alloy.

    """

    return [
        random_alloy(
            min_elements,
            max_elements,
            percentage_constraints,
            percentage_step,
            allowed_elements,
            constrain_alloy=constrain_alloys,
            structure=structures,
        )
        for _ in range(num_alloys)
    ]


def mixture(alloys: List[mg.Alloy], weights: Optional[list] = None):
    """Mix some alloys.

    :group: alloy.generate

    Parameters
    ----------

    alloys
        The alloys to be mixed.
    weights
        The weighting applied to the mixing of alloys.

    """

    for i in range(len(alloys)):
        if not isinstance(alloys[i], mg.Alloy):
            alloys[i] = mg.Alloy(alloys[i])

    if len(alloys) == 1:
        return alloys[0]

    shared_composition_space = []
    structures = []
    for alloy in alloys:
        for element in alloy.elements:
            if element not in shared_composition_space:
                shared_composition_space.append(element)
        structures.append(alloy.structure)

    constraints = None
    for alloy in alloys:
        if alloy.constraints is not None:
            if constraints is None:
                constraints = {
                    "percentages": {},
                    "min_elements": 1,
                    "max_elements": 1,
                    "percentage_step": 0.01,
                }

            if "min_elements" in alloy.constraints:
                constraints["min_elements"] = max(
                    constraints["min_elements"],
                    alloy.constraints["min_elements"],
                )
            if "max_elements" in alloy.constraints:
                constraints["max_elements"] = max(
                    constraints["max_elements"],
                    alloy.constraints["max_elements"],
                )
            if "percentage_step" in alloy.constraints:
                constraints["percentage_step"] = min(
                    constraints["percentage_step"],
                    alloy.constraints["percentage_step"],
                )

            if "percentages" in alloy.constraints:
                for element in alloy.constraints["percentages"]:
                    if element not in constraints:
                        constraints["percentages"][element] = {}

                    if "min" in alloy.constraints["percentages"][element]:
                        if "min" not in constraints["percentages"][element]:
                            constraints["percentages"][element][
                                "min"
                            ] = alloy.constraints["percentages"][element][
                                "min"
                            ]
                        else:
                            constraints["percentages"][element]["min"] = min(
                                constraints["percentages"][element]["min"],
                                alloy.constraints["percentages"][element][
                                    "min"
                                ],
                            )

                    if "max" in alloy.constraints["percentages"][element]:
                        if "max" not in constraints["percentages"][element]:
                            constraints["percentages"][element][
                                "max"
                            ] = alloy.constraints["percentages"][element][
                                "max"
                            ]
                        else:
                            constraints["percentages"][element]["max"] = max(
                                constraints["percentages"][element]["max"],
                                alloy.constraints["percentages"][element][
                                    "max"
                                ],
                            )

                    if (
                        "precedence"
                        in alloy.constraints["percentages"][element]
                    ):
                        if (
                            "precedence"
                            not in constraints["percentages"][element]
                        ):
                            constraints["percentages"][element][
                                "precedence"
                            ] = alloy.constraints["percentages"][element][
                                "precedence"
                            ]
                        else:
                            constraints["percentages"][element][
                                "precedence"
                            ] = max(
                                constraints["percentages"][element][
                                    "precedence"
                                ],
                                alloy.constraints["percentages"][element][
                                    "precedence"
                                ],
                            )

    if weights is None:
        weights = [1.0 / len(alloys)] * len(alloys)

    mixed_composition = {}
    for element in shared_composition_space:
        mixed_composition[element] = 0
        for i in range(len(alloys)):
            if element in alloys[i].elements:
                mixed_composition[element] += (
                    alloys[i].composition[element] * weights[i]
                )

    structure = [
        s
        for _, s in sorted(zip(weights, structures), key=lambda pair: pair[0])
    ][0]

    return mg.Alloy(
        mixed_composition, structure=structure, constraints=constraints
    )


def system(
    elements: Union[list, str],
    step: Number = 1,
    min_percent: Number = 0,
    max_percent: Number = 100,
    property_name: Optional[str] = None,
    crystal_structure: Optional[str] = None,
):
    """Generate a set of alloys in a particular elemental composition-space.

    :group: alloy.generate

    Parameters
    ----------

    elements
        The elements of the composition-space.
    step
        The percentage step between alloys in the composition-space.
    min_percent
        The minimum percentage of each element in alloys in the composition-space.
    max_percent
        The maximum percentage of each element in alloys in the
        composition-space.
    property_name
        A property to calculate for all alloys in the set generated.

    """

    if isinstance(elements, str):
        elements = re.findall("[A-Z][^A-Z]*", elements)

    if len(elements) == 2:
        return binary(
            elements, step, property_name, crystal_structure=crystal_structure
        )
    elif len(elements) == 3:
        return ternary(
            elements,
            step,
            min_percent,
            max_percent,
            property_name,
            crystal_structure=crystal_structure,
        )
    elif len(elements) == 4:
        return quaternary(
            elements,
            step,
            min_percent,
            max_percent,
            property_name,
            crystal_structure=crystal_structure,
        )


def binary(
    elements: Union[list, str],
    step: Number = 0.5,
    property_name: Optional[str] = None,
    crystal_structure: Optional[str] = None,
):
    """Generate a set of binary alloys in a particular elemental composition-space.

    :group: alloy.generate

    Parameters
    ----------

    elements
        The elements of the composition-space.
    step
        The percentage step between alloys in the composition-space.
    property_name
        A property to calculate for all alloys in the set generated.

    """

    if isinstance(elements, str):
        elements = re.findall("[A-Z][^A-Z]*", elements)
    if len(elements) != 2:
        raise ValueError("Binary system must have 2 elements")

    x = 100
    alloys = []
    percentages = []
    while x >= 0:
        alloy_str = elements[0] + str(x) + elements[1] + str(100 - x)
        if crystal_structure is not None:
            alloy_str += "[" + crystal_structure + "]"
        alloys.append(mg.Alloy(alloy_str))
        percentages.append(x)
        x -= step

    if property_name is not None:
        values = mg.calculate(alloys, property_name)
        return alloys, percentages, values

    return alloys, percentages


def ternary(
    elements: Union[list, str],
    step: Number = 1,
    min_percent: Number = 0,
    max_percent: Number = 100,
    property_name: Optional[str] = None,
    quaternary_element: Optional[tuple] = None,
    crystal_structure: Optional[str] = None,
):
    """Generate a set of ternary alloys.

    :group: alloy.generate

    Parameters
    ----------

    elements
        The elements of the composition-space.
    step
        The percentage step between alloys in the composition-space.
    min_percent
        The minimum percentage of each element in alloys in the composition-space.
    max_percent
        The maximum percentage of each element in alloys in the
        composition-space.
    property_name
        A property to calculate for all alloys in the set generated.
    quaternary_element
        If this ternary system is a member of an encompassing quaternary system
        provide the quaternary element and its percentage.

    """

    if isinstance(elements, str):
        elements = re.findall("[A-Z][^A-Z]*", elements)
    if len(elements) != 3:
        raise ValueError("Ternary system must have 3 elements")

    alloys = []
    percentages = []

    tmp_percentages = [max_percent + step] * 3
    while tmp_percentages[0] >= min_percent + step:
        tmp_percentages[0] -= step
        tmp_percentages[1] = max_percent + step
        tmp_percentages[2] = max_percent + step

        while tmp_percentages[1] >= min_percent + step:
            tmp_percentages[1] -= step
            tmp_percentages[2] = max_percent + step

            while tmp_percentages[2] >= min_percent + step:
                tmp_percentages[2] -= step

                if sum(tmp_percentages) == 100:

                    composition_str = ""
                    for i in range(len(elements)):
                        if tmp_percentages[i] > 0:
                            composition_str += elements[i] + str(
                                tmp_percentages[i]
                            )

                    if quaternary_element is not None:
                        composition_str = (
                            "("
                            + composition_str
                            + ")"
                            + str(100 - quaternary_element[1])
                            + quaternary_element[0]
                            + str(quaternary_element[1])
                        )

                    if crystal_structure is not None:
                        composition_str += "[" + crystal_structure + "]"

                    alloy = mg.Alloy(composition_str)
                    alloys.append(alloy)
                    percentages.append(list(alloy.composition.values()))

    if property_name is not None:
        values = mg.calculate(alloys, property_name)
        return alloys, percentages, values

    return alloys, percentages


def quaternary(
    elements: Union[list, str],
    step: Number = 1,
    min_percent: Number = 0,
    max_percent: Number = 100,
    min_quaternary_percent: Number = 0,
    quaternary_percent_step: Number = 25,
    quaternary_percentages: Optional[List[Number]] = None,
    property_name: Optional[str] = None,
    crystal_structure: Optional[str] = None,
):
    """Generate a set of quaternary alloys.

    :group: alloy.generate

    Parameters
    ----------

    elements
        The elements of the composition-space.
    step
        The percentage step between alloys in the composition-space.
    min_percent
        The minimum percentage of each element in alloys in the composition-space.
    max_percent
        The maximum percentage of each element in alloys in the
        composition-space.
    property_name
        A property to calculate for all alloys in the set generated.

    """

    if isinstance(elements, str):
        elements = re.findall("[A-Z][^A-Z]*", elements)
    if len(elements) != 4:
        raise ValueError("Quaternary system must have 4 elements")

    alloys = []

    if quaternary_percentages is None:
        quaternary_percentages = []
        quaternary_percentage = 0
        quaternary_step = 25
        while quaternary_percentage < 100:
            quaternary_percentages.append(quaternary_percentage)
            quaternary_percentage += quaternary_step

    for quaternary_percentage in quaternary_percentages:
        ternary_alloys, ternary_percentages = ternary(
            elements[:3],
            quaternary_element=(elements[3], quaternary_percentage),
            crystal_structure=crystal_structure,
        )
        alloys.append(ternary_alloys)

    return alloys



"""Module providing an interface to all other calculation functions provided by
metallurgy.
"""


import inspect
from dataclasses import fields
from numbers import Number
from typing import Callable, Iterable, List, Optional, Union

import elementy
import numpy as np

import metallurgy as mg


def is_mod_function(mod, func):
    """checks that func is a function defined in module mod"""
    return inspect.isfunction(func) and inspect.getmodule(func) == mod


def list_functions(mod):
    """list of functions defined in module mod"""
    return [
        func.__name__
        for func in mod.__dict__.values()
        if is_mod_function(mod, func)
    ]


def get_property_function(property_name: str) -> Union[Callable, None]:
    """Get the function responsible for calculating a particular property, if it
    exists in metallurgy.

    :group: utils

    Parameters
    ----------

    property_name
        Name of the property to get the function for.

    """

    # Get all modules in metallurgy
    modules = inspect.getmembers(mg, inspect.ismodule)
    for module in modules:
        # For each function in the module, check if the name matches the property
        for func in list_functions(module[1]):
            if func == property_name:
                return getattr(module[1], func)


def get_all_elemental_properties(add_suffixes: bool = False) -> List[str]:
    properties = []

    for p in fields(elementy.element.Element):
        if p.name in ["name", "symbol"]:
            continue
        if (
            p.type in [int, float, Optional[float]]
            and p.name not in properties
        ):
            if not add_suffixes:
                properties.append(p.name)
            else:
                properties.append(p.name + "_linearmix")
                properties.append(p.name + "_deviation")
                properties.append(p.name + "_range")
                properties.append(p.name + "_minimum")
                properties.append(p.name + "_maximum")

    return properties


def get_all_complex_properties() -> List[str]:
    properties = []

    # Get all modules in metallurgy
    modules = inspect.getmembers(mg, inspect.ismodule)
    for module in modules:
        for func in list_functions(module[1]):
            argspec = inspect.getfullargspec(getattr(module[1], func))
            if (
                "alloy" in argspec.args
                and len(argspec.args) == 1
                and func not in properties
            ):
                properties.append(func)
    return properties


def get_all_properties(add_suffixes: bool = False) -> List[str]:
    """Returns every calculatable property for alloy compositions.

    :group: utils
    """

    properties = (
        get_all_elemental_properties(add_suffixes=add_suffixes)
        + get_all_complex_properties()
    )

    return properties


def get_per_element_values(alloy, property_name):
    per_element_values = []
    for element in alloy.elements:
        # Get the current element's value of the property from elementy
        value = mg.periodic_table.elements[element][property_name]

        # Return None if no data for the element
        if value is None:
            return None
        # Use the first entry for the property if multiple
        elif isinstance(value, list):
            value = value[0]

        # If not numerical data, return None
        if not isinstance(value, Number):
            return None

        per_element_values.append(value)
    return per_element_values


def calculate(
    alloy: Union[mg.Alloy, str, dict],
    property_name: Optional[Union[str, List[str]]] = None,
    uncertainty: bool = False,
) -> Union[float, None, List[Union[float, None]]]:
    """Returns the a particular property calculated for an alloy, using other
    calculation functions provided by metallurgy. The property_name must match a
    function name in a metallurgy module, or be suffixed with _linearmix or
    _deviation to denote linear mixture or deviation of an elemental property.

    :group: calculations

    Parameters
    ----------

    alloy
        The alloy for which to calculate a property.
    property_name
        The property to calculate for the alloy.
    uncertainty
        If using a cerebral model, activate dropout layers during inference and
        gather uncertainty information.
    """
    if property_name is None:
        property_name = [p for p in get_all_properties(add_suffixes=True)]

    if isinstance(property_name, Iterable) and not isinstance(
        property_name, (str, dict)
    ):
        model_properties = []
        analytical_properties = []
        if mg.get_model() is not None:
            import cerebral as cb

            model = mg.get_model()
            for p in property_name:
                if p in [
                    f["name"]
                    for f in cb.models.get_model_prediction_features(model)
                ]:
                    model_properties.append(p)

        for p in property_name:
            if p not in model_properties:
                analytical_properties.append(p)

        values = {}
        for p in analytical_properties:
            values[p] = calculate(alloy, p)

        if len(model_properties) > 0:
            predictions = cb.models.predict(
                model, alloy, uncertainty=uncertainty
            )
            for p in model_properties:
                values[p] = predictions[p]

        return values

    # If we have a predictive model available to use
    if mg.get_model() is not None:
        import cerebral as cb

        # Check if the property is an output of the model, if so use the model
        # and return the prediction
        model = mg.get_model()
        if property_name in [
            f["name"] for f in cb.models.get_model_prediction_features(model)
        ]:
            return cb.models.predict(model, alloy, uncertainty=uncertainty)[
                property_name
            ]

    # Check for simple linear mixture or deviations of elemental properties
    if "_linearmix" in property_name:
        return mg.linear_mixture(alloy, property_name.split("_linearmix")[0])
    elif "_deviation" in property_name:
        return mg.deviation(alloy, property_name.split("_deviation")[0])
    elif "_range" in property_name:
        return mg.range(alloy, property_name.split("_range")[0])
    elif "_maximum" in property_name:
        return mg.maximum(alloy, property_name.split("_maximum")[0])
    elif "_minimum" in property_name:
        return mg.minimum(alloy, property_name.split("_minimum")[0])

    # Otherwise, check all function names to find a match to the property
    property_function = get_property_function(property_name)
    if property_function is not None:
        return property_function(alloy)

    # If all else fails, try a simple linear mixture again
    return mg.linear_mixture(alloy, property_name)


def linear_mixture(
    alloy: Union[mg.Alloy, str, dict], property_name: str
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the linear mixture of a particular elemental property of an
    alloy.

    See equation 5 of the paper "Machine-learning improves understanding of
    glass formation in metallic systems" for definition of the linear mixture:
    https://pubs.rsc.org/en/content/articlelanding/2022/dd/d2dd00026a

    :group: calculations

    Parameters
    ----------

    alloy : Alloy, str, dict
        The alloy for which to calculate the linear mixture.
    property_name : str
        The elemental property to calcualate the linear mixture of.

    """

    # If a list of alloys is given, return a list of linear mixture data
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [linear_mixture(a, property_name) for a in list(alloy)]

    # Convert input alloy to an Alloy instance if not already
    elif not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    # Calculate the linear mixture
    mixed_property = 0.0
    for element in alloy.elements:
        # Get the current element's value of the property from elementy
        value = mg.periodic_table.elements[element][property_name]

        # Return None if no data for the element
        if value is None:
            return None
        # Use the first entry for the property if multiple
        elif isinstance(value, list):
            value = value[0]

        # If not numerical data, return None
        if not isinstance(value, Number):
            return None

        # Element's contribution to the linear mixture, weighted by composition
        mixed_property += alloy.composition[element] * value

    return mixed_property


def deviation(
    alloy: Union[mg.Alloy, str, dict], property_name: str
) -> Union[float, None, List[Union[float, None]]]:
    """Returns the deviation of a particular elemental property in an
    alloy.

    See equation 6 of the paper "Machine-learning improves understanding of
    glass formation in metallic systems" for definition of the deviation:
    https://pubs.rsc.org/en/content/articlelanding/2022/dd/d2dd00026a

    :group: calculations

    Parameters
    ----------

    alloy
        The alloy for which to calculate the deviation.
    property_name
        The elemental property to calcualate the linear mixture of.

    """

    # If a list of alloys is given, return a list of deviation data
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [deviation(a, property_name) for a in list(alloy)]

    # Convert input alloy to an Alloy instance if not already
    elif not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    # Deviation only makes sense for multi-element alloys
    if len(alloy.elements) > 1:

        # If property is numerical, calculate the deviation of the values
        if isinstance(
            mg.periodic_table.elements[alloy.elements[0]][property_name],
            (Number, list),
        ):

            # Calculate the mean value of the property in the alloy
            mean = 0
            for element in alloy.elements:
                value = mg.periodic_table.elements[element][property_name]

                # Return None if property is None
                if value is None:
                    return None

                # Take first entry if property is a list
                if isinstance(value, list):
                    value = value[0]

                # Return None if property is not numerical
                if not isinstance(value, Number):
                    return None

                mean += alloy.composition[element] * value

            # Calculate the deviation of the property in the alloy
            total_deviation = 0
            for element in alloy.elements:
                value = mg.periodic_table.elements[element][property_name]

                # Return None if property is None
                if value is None:
                    return None

                # Take first entry if property is a list
                if isinstance(value, list):
                    value = value[0]

                # Return None if property is not numerical
                if not isinstance(value, Number):
                    return None

                total_deviation += alloy.composition[element] * (
                    (value - mean) ** 2
                )

            return total_deviation**0.5

        # If property is non-numerical, calculate the shannon entropy of the
        # values
        else:

            # Count unique values of the property
            value_count = {}
            for element in alloy.elements:
                value = mg.periodic_table.elements[element][property_name]
                if value is None:
                    return None

                if value not in value_count:
                    value_count[value] = 0
                value_count[value] += alloy.composition[element]

            # Shannon entropy is only non-zero if multiple unique values
            if len(value_count) > 1:
                shannonEntropy = 0
                for value in value_count:
                    shannonEntropy -= value_count[value] * np.log(
                        value_count[value]
                    )
                return shannonEntropy
            else:
                return 0.0

    # Single element alloys have zero deviation
    else:
        return 0.0


def range(
    alloy: Union[mg.Alloy, str, dict], property_name: str
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the range of a particular elemental property of an
    alloy.

    :group: calculations

    Parameters
    ----------

    alloy : Alloy, str, dict
        The alloy for which to calculate the range.
    property_name : str
        The elemental property to calcualate the linear mixture of.

    """

    # If a list of alloys is given, return a list of linear mixture data
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mg.range(a, property_name) for a in list(alloy)]

    # Convert input alloy to an Alloy instance if not already
    elif not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    per_element_values = get_per_element_values(alloy, property_name)
    if per_element_values is None or None in per_element_values:
        return None
    return float(
        np.abs(np.max(per_element_values) - np.min(per_element_values))
    )


def maximum(
    alloy: Union[mg.Alloy, str, dict], property_name: str
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the maximum of a particular elemental property of an alloy.

    :group: calculations

    Parameters
    ----------

    alloy : Alloy, str, dict
        The alloy for which to calculate the maximum.
    property_name : str
        The elemental property to calcualate the linear mixture of.

    """

    # If a list of alloys is given, return a list of linear mixture data
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [maximum(a, property_name) for a in list(alloy)]

    # Convert input alloy to an Alloy instance if not already
    elif not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    raw_values = get_per_element_values(alloy, property_name)
    if raw_values is not None:
        values = []
        for v in raw_values:
            if v is not None:
                values.append(v)

        return float(np.max(values))


def minimum(
    alloy: Union[mg.Alloy, str, dict], property_name: str
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the minimum of a particular elemental property of an alloy.

    :group: calculations

    Parameters
    ----------

    alloy : Alloy, str, dict
        The alloy for which to calculate the minimum.
    property_name : str
        The elemental property to calcualate the linear mixture of.

    """

    # If a list of alloys is given, return a list of linear mixture data
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [minimum(a, property_name) for a in list(alloy)]

    # Convert input alloy to an Alloy instance if not already
    elif not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    raw_values = get_per_element_values(alloy, property_name)
    if raw_values is not None:
        values = []
        for v in raw_values:
            if v is not None:
                values.append(v)
        return float(np.min(values))



from __future__ import annotations

import copy
import math
import re
from collections import Counter, OrderedDict
from types import SimpleNamespace
from typing import Callable, Optional, Union

import elementy
import numpy as np

from .prototype import Prototype
from .prototypes import get_prototype


class Alloy:
    """An alloy, a mixture of chemical elements with specific percentages.

    :group: alloy

    Attributes
    ----------

    composition
        Dictionary matching element symbols to atomic percentages.
    structure
        String referring to a crystal structure prototype which defines atomic
        percentages.
    constraints
        Dictionary of constraints to follow when adjusting atomic  percentages.

    """

    class Composition(OrderedDict):
        """Atomic percentages of elements in an alloy.

        :group: alloy

        Attributes
        ----------

        on_change : Callable
            Callback function enabling parent to act on composition changes

        """

        def __init__(self, value: dict, on_change: Callable, *args, **kwargs):
            super().__init__(value, *args, **kwargs)
            self.on_change = on_change

        def __setitem__(
            self, element: str, percentage: float, respond_to_change=True
        ):
            """
            Set the percentage value of an element in the composition.
            Remove elements from the composition if below minimum threshold.
            """
            if percentage >= 0.0001:
                super().__setitem__(element, percentage)
            else:
                if element in self.keys():
                    super().__delitem__(element)
            if hasattr(self, "on_change") and respond_to_change:
                self.on_change()

        def __delitem__(self, element, respond_to_change=True):
            super().__delitem__(element)
            if hasattr(self, "on_change") and respond_to_change:
                self.on_change()

    def __init__(
        self,
        composition: Union[str, dict, Alloy],
        structure: Optional[Union[str, Prototype]] = None,
        constraints: Optional[dict] = None,
        rescale: bool = True,
    ):

        self.original_composition, composition_structure = parse_composition(
            composition
        )
        self.original_elements = list(self.original_composition.keys())
        self.composition = filter_order_composition(self.original_composition)

        if self.composition is None:
            raise Exception("Invalid composition:", composition)
        if structure is None and composition_structure is not None:
            structure = composition_structure

        self.constraints = None

        self._structure = None
        if structure is not None:
            self.structure = structure

        if rescale:
            self.rescale()

        if constraints is not None:
            self.constraints = parse_constraints(**constraints)
            if self.num_elements == 1:
                if self.elements[0] in self.constraints["percentages"]:
                    self.constraints["percentages"][self.elements[0]][
                        "max"
                    ] = 1
            if rescale:
                self.rescale()

    def __repr__(self):
        return self.to_string()

    def __eq__(self, other):
        if isinstance(other, Alloy):
            return self.to_string() == other.to_string()
        elif isinstance(other, (str, dict)):
            return self.to_string() == Alloy(other).to_string()
        else:
            return False

    def __add__(self, other):
        from . import generate

        return generate.mixture([self, other])

    def __hash__(self):
        return hash(self.to_string())

    @property
    def composition(self) -> Composition:
        """Dictionary of elements and percentages in the alloy.

        :group: alloy
        """
        return self._composition

    @composition.setter
    def composition(self, value):
        if isinstance(value, (dict, OrderedDict)):
            value = self.Composition(value, self.on_composition_change)
        self._composition = value

    def on_composition_change(self):
        """Called when composition property changes.

        :group: alloy.utils
        """
        self.determine_percentage_constraints()

        if not hasattr(self, "rescaling"):
            self.rescale()

    @property
    def structure(self) -> str:
        return self._structure

    @structure.setter
    def structure(self, structure):

        if isinstance(structure, str):
            try:
                structure = copy.deepcopy(get_prototype(structure))
            except:
                self._structure = SimpleNamespace(**{"name": structure})
                return

        self._structure = structure

        structure_elements_ordered = sorted(
            self._structure.composition,
            key=lambda x: self._structure.composition[x],
            reverse=True,
        )

        for i in range(len(self.elements)):
            if i not in structure_elements_ordered:
                structure_elements_ordered.append(i)

        if not all(
            value == list(self.original_composition.values())[0]
            for value in self.original_composition.values()
        ):
            digits = 0
            for element in self.composition:
                digits = max([digits, len(str(self.composition[element])) - 2])
            digits = min([5, digits])

        self.structure_element_translation = {}
        for i, element in enumerate(self.elements):
            if i >= len(self.structure.elements):
                structure_element_index = i % len(self.structure.elements)
            else:
                structure_element_index = i
            self.structure_element_translation[i] = structure_element_index

        if len(self.elements) < len(self.structure.elements):
            for i, element in enumerate(self.structure.elements):
                if i not in self.structure_element_translation:
                    self.structure_element_translation[i] = i % len(
                        self.elements
                    )

        structure_composition = {}
        for i in self.structure_element_translation:
            element_index = self.structure_element_translation[i]
            if i < len(self.elements):
                if i not in structure_composition:
                    structure_composition[i] = 0
                structure_composition[i] += self.structure.composition[
                    element_index
                ]
            else:
                if element_index not in structure_composition:
                    structure_composition[element_index] = 0
                structure_composition[
                    element_index
                ] += self.structure.composition[i]

        if not all(
            value == list(self.original_composition.values())[0]
            for value in self.original_composition.values()
        ):
            total = sum(structure_composition.values())
            for i in structure_composition:
                structure_composition[i] /= total
                structure_composition[i] = multiple_round(
                    structure_composition[i], 10 ** (-digits - 1), digits + 1
                )

            self.structure_element_translation = {}
            taken = []

            for element in self.composition:
                for i in structure_composition:
                    if i not in taken and math.isclose(
                        self.composition[element],
                        structure_composition[i],
                        abs_tol=10 ** (-digits),
                    ):
                        self.structure_element_translation[element] = i
                        taken.append(i)
                        break

            translated_structure_composition = {}
            for element in self.structure_element_translation:
                translated_structure_composition[
                    element
                ] = structure_composition[
                    self.structure_element_translation[element]
                ]

            if self.constraints is None:
                self.constraints = {
                    "digits": digits,
                    "percentage_step": (10**-digits),
                }
        else:
            translated_structure_composition = {}
            for i in self.structure_element_translation:
                if i >= self.num_elements:
                    continue
                translated_structure_composition[
                    self.elements[i]
                ] = structure_composition[i]

            filtered_translation = {}
            for i in structure_composition:
                filtered_translation[
                    self.elements[i]
                ] = self.structure_element_translation[i]

            self.structure_element_translation = filtered_translation

        if len(translated_structure_composition) > 0:
            self.composition = translated_structure_composition

        self.rescale()

    @property
    def elements(self) -> list:
        """List of elements in the alloy.

        :group: alloy
        """
        return list(self.composition.keys())

    @property
    def elements_structure_order(self) -> list:
        """List of elements in the alloy in the order of the structure basis.

        :group: alloy
        """
        if self.structure is not None:
            sorted_translation = sorted(
                self.structure_element_translation.items(),
                key=lambda x: x[1],
            )

            elements = []
            for pair in sorted_translation:
                elements.append(pair[0])
            return elements

        else:
            return self.elements

    @property
    def num_elements(self) -> int:
        """Number of elements in the alloy.

        :group: alloy
        """
        return len(self.elements)

    @property
    def total_percentage(self) -> float:
        """Sum of percentages in the alloy composition.

        :group: alloy
        """
        return sum(self.composition.values())

    def add_element(self, element: str, percentage: Optional[float] = 0.0):
        """Adds an element to the alloy composition.

        :group: alloy

        Parameters
        ----------

        element
            Element to be added.
        percentage
            Percentage of the new element. If zero or not set, is given the
            value of 1.0/(number of elements +1).

        """
        if element not in self.elements:
            if percentage == 0.0:
                percentage = 1.0 / (self.num_elements + 1)

            self.composition[element] = percentage

    def remove_element(self, element: str):
        """Removes an element from the alloy composition.

        :group: alloy

        Parameters
        ----------

        element
            Element to be removed.

        """
        if element in self.elements:
            del self.composition[element]

    def ensure_constrained_elements_present(self):
        """Adds to the composition any elements with minimum percentage
        constraints that are missing

        :group: alloy.utils
        """
        if "local_percentages" in self.constraints:
            for element in self.constraints["local_percentages"]:
                if element not in self.composition:
                    if (
                        self.constraints["local_percentages"][element]["min"]
                        > 0
                    ):
                        self.composition[element] = self.constraints[
                            "local_percentages"
                        ][element]["min"]

    def rescale(self):
        """Adjust elemental percentages to match constraints

        All alloys are constrained such that element percentages sum
        to 1.0. Additional constraints include:

        - Minimum number of unique elements in an alloy.
        - Maximum number of unique elements in an alloy.
        - Maximum and minimum percentage values per element.
        - Precedence order of elements.

        :group: alloy.utils
        """

        if len(self.composition) < 1:
            return

        self.rescaling = True

        constraints_applied = False
        while not self.constraints_satisfied():

            self.determine_percentage_constraints()

            constraints_applied = True

            precedence = {}
            no_precedence = []

            self.ensure_constrained_elements_present()
            for element in self.composition:
                if element in self.constraints["local_percentages"]:
                    if (
                        self.constraints["local_percentages"][element]["min"]
                        > 0
                    ):
                        precedence[element] = self.constraints[
                            "local_percentages"
                        ][element]["precedence"]

                if element not in precedence:
                    no_precedence.append(element)

            precedence_order = (
                sorted(precedence, key=precedence.get, reverse=True)
                + no_precedence
            )
            reverse_precedence_order = precedence_order[::-1]

            for element in precedence_order:
                if element not in self.constraints["local_percentages"]:
                    continue

                for direction in ["min", "max"]:
                    if element_is_in_range(
                        element,
                        self.composition,
                        self.constraints,
                        direction,
                        inclusive=True,
                    ):
                        continue

                    for other_element in reverse_precedence_order:
                        if (
                            element != other_element
                            and other_element in self.composition
                        ):
                            if not element_is_in_range(
                                other_element,
                                self.composition,
                                self.constraints,
                                direction,
                                inclusive=False,
                            ):
                                continue

                            if direction == "min":
                                donate_percentage(
                                    self.composition,
                                    self.constraints,
                                    other_element,
                                    element,
                                )
                            else:
                                donate_percentage(
                                    self.composition,
                                    self.constraints,
                                    element,
                                    other_element,
                                )

                            self.ensure_constrained_elements_present()

                            if element_is_in_range(
                                element,
                                self.composition,
                                self.constraints,
                                direction,
                            ):
                                break

            self.constrain_num_elements()
            self.clamp_composition()
            self.round_composition()

        self.reorder_composition()

        if not constraints_applied:
            self.clamp_composition()
        self.round_composition()

        delattr(self, "rescaling")

    def reorder_composition(self):
        ordered_elements = self.elements
        if self.constraints is not None and "percentages" in self.constraints:
            ordered_elements = sorted(
                self.composition,
                key=lambda x: (
                    self.constraints["percentages"][x]["precedence"]
                    if x in self.constraints["percentages"]
                    else 0
                ),
                reverse=True,
            )
        ordered_elements = sorted(
            ordered_elements,
            key=self.composition.get,
            reverse=True,
        )
        for element in ordered_elements:
            self.composition.move_to_end(element)

    def constrain_max_elements(self):
        """Removes elements from an alloy if there are more elements than
        allowed by the max_elements constraint."""
        if len(self.composition) > self.constraints["max_elements"]:

            excess = len(self.composition) - self.constraints["max_elements"]

            toDelete = []
            ordered_elements = sorted(
                self.composition, key=self.composition.get
            )
            for element in ordered_elements:
                if element in self.constraints["local_percentages"]:
                    if (
                        self.constraints["local_percentages"][element]["min"]
                        == 0
                    ):
                        toDelete.append(element)
                else:
                    toDelete.append(element)

                if len(toDelete) == excess:
                    break

            for element in toDelete:
                del self.composition[element]

    def constrain_min_elements(self):
        """Adds elements to an alloy if there are fewer elements than
        allowed by the min_elements constraint."""

        while len(self.composition) < self.constraints["min_elements"]:
            element_to_add = np.random.choice(
                self.constraints["allowed_elements"], 1
            )[0]
            if element_to_add not in self.composition:
                self.composition.__setitem__(
                    element_to_add, np.random.uniform()
                )

    def constrain_num_elements(self):
        """Applies max and min element count constraints.
        :group: alloy.utils
        """
        self.constrain_max_elements()
        self.constrain_min_elements()

    def constraints_satisfied(self) -> bool:
        """Check for satisfaction of the constraints acting on an alloy.

        :group: alloy.utils
        """

        satisfied = True
        if (
            self.constraints is not None
            and "local_percentages" in self.constraints
        ):

            discrepancy = np.abs(
                1
                - multiple_round(
                    self.total_percentage,
                    self.constraints["percentage_step"],
                    self.constraints["digits"],
                )
            )

            if discrepancy > self.constraints["percentage_step"]:
                # print("Sum != 1", self.total_percentage, discrepancy)
                # print(self.composition)
                # print()
                satisfied = False

            if satisfied and "local_percentages" in self.constraints:

                for element in self.constraints["local_percentages"]:
                    if element in self.composition:
                        if (
                            self.composition[element]
                            < self.constraints["local_percentages"][element][
                                "min"
                            ]
                        ):
                            # print(
                            #     "element below min",
                            #     element,
                            #     self.composition[element],
                            #     self.constraints["local_percentages"][element][
                            #         "min"
                            #     ],
                            # )
                            # print(self.composition)
                            # print()
                            satisfied = False
                            break
                        elif (
                            self.composition[element]
                            > self.constraints["local_percentages"][element][
                                "max"
                            ]
                        ):
                            # print(
                            #     "element above max",
                            #     element,
                            #     self.composition[element],
                            #     self.constraints["local_percentages"][element][
                            #         "max"
                            #     ],
                            # )
                            # print(self.composition)
                            # print()
                            satisfied = False
                            break
                        elif (
                            "superior_elements"
                            in self.constraints["local_percentages"][element]
                        ):
                            for superior_element in self.constraints[
                                "local_percentages"
                            ][element]["superior_elements"]:
                                if superior_element in self.composition:
                                    if (
                                        self.composition[element]
                                        > self.composition[superior_element]
                                    ):
                                        # print(
                                        #     "element above superior",
                                        #     element,
                                        #     superior_element,
                                        #     self.composition[element],
                                        #     self.constraints[
                                        #         "local_percentages"
                                        #     ][element]["max"],
                                        #     self.composition[superior_element],
                                        # )
                                        # print(self.composition)
                                        # print(self.constraints)
                                        # print()
                                        satisfied = False
                                        break
                                if not satisfied:
                                    break

                    elif (
                        self.constraints["local_percentages"][element]["min"]
                        > 0.0
                    ):
                        # print(
                        #     "element missing",
                        #     element,
                        #     self.constraints["local_percentages"][element][
                        #         "min"
                        #     ],
                        # )
                        # print(self.composition)
                        # print()
                        satisfied = False
                        break

            if satisfied:
                if (
                    "max_elements" in self.constraints
                    and len(self.composition)
                    > self.constraints["max_elements"]
                ):
                    satisfied = False
                    # print("too many elements")
                    # print(self.composition)
                    # print()
                elif (
                    "min_elements" in self.constraints
                    and len(self.composition)
                    < self.constraints["min_elements"]
                ):
                    satisfied = False
                    # print("too few elements")
                    # print(self.composition)
                    # print()

        return satisfied

    def clamp_composition(self):
        """Adjust elemental percentages such that they sum to 1

        :group: alloy.utils
        """

        digits = 4
        percentage_step = 0.01
        if self.constraints is not None:
            if "digits" in self.constraints:
                digits = self.constraints["digits"]
            if "percentage_step" in self.constraints:
                percentage_step = self.constraints["percentage_step"]

        while (
            np.abs(
                1
                - multiple_round(
                    self.total_percentage, percentage_step, digits
                )
            )
            >= percentage_step
        ):

            current_total = self.total_percentage
            elements = self.elements

            clamped_rounded_total = 0
            for element in elements:
                clamped_value = self.composition[element] / current_total
                clamped_value = multiple_round(
                    clamped_value, percentage_step, digits
                )

                if (
                    self.constraints is not None
                    and "percentages" in self.constraints
                ):
                    if element in self.constraints["percentages"]:
                        clamped_value = max(
                            clamped_value,
                            self.constraints["percentages"][element]["min"],
                        )

                self.composition.__setitem__(
                    element, clamped_value, respond_to_change=False
                )

                if element in self.composition:
                    clamped_rounded_total += self.composition[element]

            while (
                round(np.abs(1 - clamped_rounded_total), digits)
                >= percentage_step
            ):
                elements = self.elements
                for element in elements:
                    if element not in self.composition:
                        continue

                    if (
                        self.constraints is not None
                        and "percentages" in self.constraints
                    ):
                        if element in self.constraints["percentages"]:
                            change = max(
                                [
                                    percentage_step,
                                    max(
                                        [
                                            0,
                                            self.constraints["percentages"][
                                                element
                                            ]["max"]
                                            - self.composition[element],
                                        ]
                                    ),
                                ]
                            )
                        else:
                            change = percentage_step
                    else:
                        change = percentage_step

                    change = min([change, np.abs(1 - clamped_rounded_total)])
                    if clamped_rounded_total > 1:
                        change *= -1

                    change = multiple_round(change, percentage_step, digits)

                    self.composition.__setitem__(
                        element,
                        self.composition[element] + change,
                        respond_to_change=False,
                    )
                    clamped_rounded_total += change
                    if np.abs(1 - clamped_rounded_total) > percentage_step / 2:
                        break

    def determine_percentage_constraints(self):
        """Determine the current constraints on an alloy composition.

        When precedence constraints are acting, the maximum and
        minimum percentage values for each element depend on the
        others.

        :group: alloy.utils
        """

        if self.constraints is None or "percentages" not in self.constraints:
            return

        tmp_percentages = copy.deepcopy(self.constraints["percentages"])

        lowest_precedence = {
            "element": None,
            "precedence": np.Inf,
            "percentage": np.Inf,
        }

        for element in self.constraints["percentages"]:
            if element in self.composition:
                element_constraint = self.constraints["percentages"][element]
                if (
                    element_constraint["precedence"]
                    <= lowest_precedence["precedence"]
                    and element_constraint["precedence"] > 0
                    and self.composition[element]
                    < lowest_precedence["percentage"]
                ):

                    lowest_precedence["precedence"] = element_constraint[
                        "precedence"
                    ]
                    lowest_precedence["element"] = element
                    lowest_precedence["percentage"] = self.composition[element]

            if (
                len(
                    self.constraints["percentages"][element][
                        "superior_elements"
                    ]
                )
                > 0
            ):
                superior_elements = []
                for e in self.constraints["percentages"][element][
                    "superior_elements"
                ]:
                    if e in self.composition:
                        superior_elements.append(self.composition[e])

                if len(superior_elements) > 0:
                    tmp_percentages[element]["max"] = min(
                        [
                            self.constraints["percentages"][element]["max"],
                            min(superior_elements),
                        ]
                    )

                    tmp_percentages[element]["min"] = min(
                        [
                            tmp_percentages[element]["min"],
                            tmp_percentages[element]["max"],
                        ]
                    )

            # if (
            #     len(
            #         self.constraints["percentages"][element][
            #             "inferior_elements"
            #         ]
            #     )
            #     > 0
            # ):
            #     inferior_elements = []
            #     for e in self.constraints["percentages"][element][
            #         "inferior_elements"
            #     ]:
            #         if e in self.composition:
            #             inferior_elements.append(self.composition[e])

            #     if len(inferior_elements) > 0:
            #         tmp_percentages[element]["min"] = max(
            #             [
            #                 self.constraints["percentages"][element]["min"],
            #                 max(inferior_elements),
            #             ]
            #         )

            #         tmp_percentages[element]["max"] = max(
            #             [
            #                 tmp_percentages[element]["min"],
            #                 tmp_percentages[element]["max"],
            #             ]
            #         )

        if lowest_precedence["element"] is not None:
            for element in self.composition:
                if element not in tmp_percentages:
                    tmp_percentages[element] = {
                        "max": lowest_precedence["percentage"],
                        "min": 0,
                        "precedence": 0,
                        "superior_elements": [],
                    }

                    for other_element in tmp_percentages:
                        if element != other_element:
                            if (
                                tmp_percentages[element]["precedence"]
                                < tmp_percentages[other_element]["precedence"]
                            ):
                                tmp_percentages[element][
                                    "superior_elements"
                                ].append(other_element)

        self.constraints["local_percentages"] = tmp_percentages

        if self.constraints["max_elements"] == len(
            self.constraints["local_percentages"]
        ):
            total_max = 0
            for element in self.composition:
                if element in self.constraints["local_percentages"]:
                    total_max += self.constraints["local_percentages"][
                        element
                    ]["max"]
            while total_max < 1:
                for element in self.constraints["local_percentages"]:
                    if element not in self.composition:
                        value = min(
                            [
                                self.constraints["local_percentages"][element][
                                    "max"
                                ],
                                1 - total_max,
                            ]
                        )
                        self.composition.__setitem__(
                            element, value, respond_to_change=False
                        )
                        total_max += value

                if total_max < 1 and len(
                    self.constraints["allowed_elements"]
                ) > len(self.composition):
                    element_to_add = np.random.choice(
                        self.constraints["allowed_elements"], 1
                    )[0]
                    if element_to_add not in self.composition:
                        value = np.random.uniform()
                        self.composition.__setitem__(
                            element_to_add,
                            value,
                            respond_to_change=False,
                        )
                        total_max += value
                else:
                    for element in self.elements:
                        value = np.random.uniform()
                        self.composition.__setitem__(
                            element,
                            self.composition[element] + value,
                            respond_to_change=False,
                        )
                        total_max += value

    def to_string(self) -> str:
        """Convert the alloy composition to a string

        :group: alloy.utils
        """

        composition_str = ""
        for element in self.composition:
            percentage_str = str(self.composition[element] * 100.0)

            split_str = percentage_str.split(".")
            decimal = split_str[1]
            if decimal == "0":
                percentage_str = split_str[0]
            else:
                decimal_places = len(
                    str(self.composition[element]).split(".")[1]
                )

                percentage_str = str(
                    round(float(percentage_str), decimal_places)
                )

                split_str = percentage_str.split(".")
                decimal = split_str[1]
                if decimal == "0":
                    percentage_str = split_str[0]

            composition_str += element + percentage_str

        if self.structure is not None:
            composition_str += "[" + self.structure.name + "]"

        return composition_str

    def to_pretty_string(self) -> str:
        """Convert alloy composition to string with LaTeX formatting of
        subscripts

        :group: alloy.utils
        """
        numbers = re.compile(r"(\d+)")
        return numbers.sub(r"$_{\1}$", self.to_string())

    def to_dict(self) -> dict:
        """Convert the alloy composition to a dictionary

        :group: alloy.utils
        """
        alloy_dict = {"composition": self.to_string()}
        return alloy_dict

    def round_composition(self):
        """Round elemental percentages in composition while maintaining sum

        :group: alloy.utils
        """

        if len(self.composition) == 0:
            return
        elif len(self.composition) == 1:
            self.composition[self.elements[0]] = 1.0

        digits = 4
        percentage_step = 0.0001
        if self.constraints is not None:
            if "digits" in self.constraints:
                digits = self.constraints["digits"]
            if "percentage_step" in self.constraints:
                percentage_step = self.constraints["percentage_step"]

        needs_rounding = False
        for element in self.elements:
            percentage_str = str(self.composition[element])
            length = len(percentage_str)
            if "." in percentage_str:
                length -= 2
            if length > digits:
                easy_round = False
                easy_rounders = ["0", "9"]
                count = Counter(percentage_str)
                for c in easy_rounders:
                    if c in count and not easy_round:
                        count[c] /= len(percentage_str)
                        if count[c] > 0.5:
                            easy_round = True
                            self.composition[element] = multiple_round(
                                self.composition[element],
                                percentage_step,
                                digits,
                            )
                if not easy_round:
                    needs_rounding = True
                    break

        if not needs_rounding:
            return

        integer_parts = []
        decimal_parts = []

        for element in self.elements:
            percentage = str(self.composition[element] * (10**digits))
            split_percentage = percentage.split(".")

            integer_parts.append(int(split_percentage[0]))

            if len(split_percentage) > 1:
                decimal_parts.append(
                    float(percentage) - float(split_percentage[0])
                )
            else:
                decimal_parts.append(0.0)

        undershoot = (10**digits) - sum(integer_parts)

        if undershoot > 0:
            if self.constraints is not None:

                precedence = {}
                for element in self.elements:
                    if (
                        "local_percentages" in self.constraints
                        and element in self.constraints["local_percentages"]
                    ):
                        precedence[element] = self.constraints[
                            "local_percentages"
                        ][element]["precedence"]
                    else:
                        precedence[element] = 0
                precedence_order = sorted(
                    precedence, key=precedence.get, reverse=True
                )

                filtered_precedence_order = []
                for element in precedence_order:
                    if element in self.composition:
                        filtered_precedence_order.append(element)
            else:
                filtered_precedence_order = self.elements

            i = 0
            while undershoot > 0:
                decimal_part_index = self.elements.index(
                    filtered_precedence_order[i]
                )

                constraints_violated = False
                if self.constraints is not None:
                    if (
                        "local_percentages" in self.constraints
                        and filtered_precedence_order[i]
                        in self.constraints["local_percentages"]
                    ):
                        if decimal_part_index < len(integer_parts):

                            if (integer_parts[decimal_part_index] + 1) / (
                                10**digits
                            ) > self.constraints["local_percentages"][
                                filtered_precedence_order[i]
                            ][
                                "max"
                            ]:
                                constraints_violated = True

                if not constraints_violated:
                    if decimal_part_index < len(integer_parts):
                        integer_parts[decimal_part_index] += 1
                        undershoot -= 1
                else:
                    break
                i += 1

                if i >= len(filtered_precedence_order):
                    i = 0

            i = 0
            for element in self.elements:
                if i < len(integer_parts):
                    self.composition[element] = multiple_round(
                        integer_parts[i] / (10**digits),
                        percentage_step,
                        digits,
                    )
                else:
                    self.composition[element] = 0
                i += 1

        else:
            i = 0
            for element in self.elements:

                if i < len(integer_parts):
                    rounded_value = multiple_round(
                        integer_parts[i] / (10**digits),
                        percentage_step,
                        digits,
                    )

                    if self.constraints is not None:
                        if element in self.constraints["local_percentages"]:
                            self.composition[element] = max(
                                self.constraints["local_percentages"][element][
                                    "min"
                                ],
                                rounded_value,
                            )
                        else:
                            self.composition[element] = rounded_value
                    else:
                        self.composition[element] = rounded_value
                else:
                    self.composition[element] = 0

                i += 1

        deviation = 1 - self.total_percentage
        abs_deviation = np.abs(deviation)
        if abs_deviation >= percentage_step:
            num_steps = int(abs_deviation // percentage_step)

            j = 0
            for _ in range(num_steps):
                step = percentage_step
                if deviation < 0:
                    step *= -1

                self.composition.__setitem__(
                    self.elements[j],
                    self.composition[self.elements[j]] + step,
                    respond_to_change=False,
                )
                if j > self.num_elements:
                    j = 0


def parse_composition(composition: Union[str, dict, Alloy]) -> dict:
    """Parse elemental percentages of an alloy from input

    :group: alloy.utils
    """
    if isinstance(composition, str):
        return parse_composition_string(composition)
    elif isinstance(composition, dict):
        return parse_composition_dict(composition), None
    elif isinstance(composition, Alloy):
        return composition.composition, None


def parse_composition_string(composition_string: str) -> dict:
    """Parse elemental percentages of an alloy from a string

    :group: alloy.utils
    """

    if len(composition_string) == 0:
        return None

    if "[" in composition_string:
        structure = composition_string.split("[")[1].split("]")[0]
        composition_string = composition_string.split("[")[0]
    else:
        structure = None

    if "(" in composition_string:

        sub_composition_string = composition_string.split(")")[0].split("(")[1]

        sub_composition = parse_composition_string_block(
            sub_composition_string
        )

        pre_sub_composition_string = composition_string.split(")")[0].split(
            "("
        )[0]
        post_sub_composition_string = composition_string.split(")")[1].split(
            "("
        )[0]

        remaining_composition_string = ""
        if re.search("[a-zA-Z]", pre_sub_composition_string) is not None:
            remaining_composition_string += pre_sub_composition_string
        if re.search("[a-zA-Z]", post_sub_composition_string) is not None:
            remaining_composition_string += post_sub_composition_string

        split_composition = re.findall(
            r"[A-Z][^A-Z]*", remaining_composition_string
        )

        composition_sum = 0
        for element_percentage in split_composition:
            split_element_percentage = re.split(
                r"(\d+(?:\.\d+)?)", element_percentage
            )
            if len(split_element_percentage) > 1:
                composition_sum += float(split_element_percentage[1])
        if composition_sum == 0:
            composition_sum = 1

        sub_composition_percentage = float(
            re.split(r"(\d+(?:\.\d+)?)", composition_string.split(")")[1])[1]
        )
        sub_composition_sum = sum(
            [sub_composition[p] for p in sub_composition]
        )
        total_sum = composition_sum + sub_composition_percentage
        scale_factor = 1
        if total_sum > 1.0:
            scale_factor = 100
        sub_composition_percentage /= scale_factor
        for element in sub_composition:
            sub_composition[element] *= (
                sub_composition_percentage / sub_composition_sum
            )

        remaining_composition = parse_composition_string_block(
            remaining_composition_string
        )
        for element in remaining_composition:
            remaining_composition[element] *= 1 - sub_composition_percentage

        composition = {**sub_composition, **remaining_composition}

    else:
        composition = parse_composition_string_block(composition_string)

    return composition, structure


def parse_composition_string_block(composition_string: str) -> dict:
    """Parse a block of a composition string, found between brackets, returning
    a composition dictionary for that block

    :group: alloy.utils
    """

    composition = {}

    split_composition = re.findall(r"[A-Z][^A-Z]*", composition_string)

    composition_sum = 0
    for element_percentage in split_composition:
        split_element_percentage = re.split(
            r"(\d+(?:\.\d+)?)", element_percentage
        )
        if len(split_element_percentage) > 1:
            composition_sum += float(split_element_percentage[1])
    if composition_sum == 0:
        composition_sum = 1

    for element_percentage in split_composition:
        split_element_percentage = re.split(
            r"(\d+(?:\.\d+)?)", element_percentage
        )

        if len(split_element_percentage) > 1:
            decimal_places = 2
            if "." in str(split_element_percentage[1]):
                decimal_places += len(
                    str(split_element_percentage[1]).split(".")[1]
                )

            composition[split_element_percentage[0]] = round(
                float(split_element_percentage[1]) / composition_sum,
                decimal_places,
            )
        else:
            composition[split_element_percentage[0]] = 1.0 / len(
                split_composition
            )

    return composition


def parse_composition_dict(composition: dict) -> dict:
    """Parse elemental percentages of an alloy from a dict

    :group: alloy.utils
    """

    if len(composition) == 0:
        return None

    needRescale = False
    for element in composition:
        if composition[element] > 1:
            needRescale = True
            break

    if needRescale:
        for element in composition:
            composition[element] /= 100.0

    return filter_order_composition(composition)


def filter_order_composition(composition: dict) -> OrderedDict:
    """Filter & order elements in an alloy composition

    Elements with percentage less than zero are filtered out.
    An OrderedDict is used to sort the composition in descending
    order of elemental percentage.

    :group: alloy.utils
    """

    filtered_composition = {}
    for element in composition:
        if composition[element] > 0:
            filtered_composition[element] = composition[element]

    ordered_composition = OrderedDict()
    for element in sorted(
        filtered_composition, key=filtered_composition.get, reverse=True
    ):
        ordered_composition[element] = filtered_composition[element]

    return ordered_composition


def parse_constraints(
    min_elements: int = 1,
    max_elements: int = 10,
    percentages: dict = {},
    allowed_elements: list = [e for e in elementy.PeriodicTable().elements],
    disallowed_elements: list = [],
    percentage_step: float = 0.01,
) -> dict:
    """Parse constraint rules from input

    :group: alloy.utils

    Parameters
    ----------
        min_elements
            Minimum number of elements in an alloy
        max_elements
            Maximum number of elements in an alloy
        percentages
            Dict of maximum and minimum percentages per element
        allowed_elements
            List of elements allowed in an alloy
        disallowed_elements
            List of elements not allowed in an alloy
        digits
            Number of significant figures to consider for percentages
        percentage_step
            Increment between percentages

    """

    if not isinstance(percentages, dict):
        if isinstance(percentages, list):
            tmp_percentages = {}
            for element in percentages:
                tmp_percentages[element] = {}
            percentages = tmp_percentages

    for element in percentages:
        if "min" not in percentages[element]:
            percentages[element]["min"] = 0.0
        else:
            percentages[element]["min"] = max(percentages[element]["min"], 0.0)

        if "max" not in percentages[element]:
            percentages[element]["max"] = 1.0
        else:
            percentages[element]["max"] = min(percentages[element]["max"], 1.0)

        if "precedence" not in percentages[element]:
            percentages[element]["precedence"] = 0

    for element in percentages:
        percentages[element]["superior_elements"] = []
        percentages[element]["inferior_elements"] = []
        for other_element in percentages:
            if element != other_element:
                if (
                    percentages[element]["precedence"]
                    < percentages[other_element]["precedence"]
                ):
                    percentages[element]["superior_elements"].append(
                        other_element
                    )
                elif (
                    percentages[element]["precedence"]
                    > percentages[other_element]["precedence"]
                ):
                    percentages[element]["inferior_elements"].append(
                        other_element
                    )

        if len(percentages[element]["superior_elements"]) > 0:
            inferior_min = percentages[element]["min"]
            for other_element in percentages:
                if (
                    other_element
                    not in percentages[element]["superior_elements"]
                ):
                    inferior_min = max(
                        [inferior_min, percentages[other_element]["min"]]
                    )
            if inferior_min < percentages[element]["max"]:
                percentages[element]["min"] = inferior_min

        if len(percentages[element]["inferior_elements"]) > 0:
            superior_min = percentages[element]["min"]
            for other_element in percentages[element]["inferior_elements"]:
                superior_min = max(
                    [superior_min, percentages[other_element]["min"]]
                )
            if superior_min > percentages[element]["min"]:
                percentages[element]["min"] = superior_min

    min_sum = sum([percentages[e]["min"] for e in percentages])
    if min_sum > 1:
        print(
            "Impossible constraints, mins for each element sum greater than 1"
        )

    for element in disallowed_elements:
        if element in allowed_elements:
            allowed_elements.remove(element)

    digits = len(str(percentage_step))
    if "." in str(percentage_step):
        digits -= 2

    return {
        "percentages": percentages,
        "local_percentages": percentages,
        "allowed_elements": allowed_elements,
        "min_elements": min_elements,
        "max_elements": max_elements,
        "digits": digits,
        "percentage_step": percentage_step,
    }


def donate_percentage(
    composition: Alloy.Composition,
    constraints: dict,
    donor: str,
    recipient: str,
):
    """Adjusts a composition by transferring some percentage from one element
    (the donor) to another element (the recipient).

    :group: alloy.utils

    Parameters
    ----------

    composition
        The percentages of the alloy.
    constraints
        The constraints on the alloy.
    donor
        The element to give away some percentage.
    recipient
        The element to recieve some percentage.

    """

    recipient_deficit = 0
    donor_surplus = 0

    if (
        recipient in constraints["local_percentages"]
        and recipient in composition
    ):
        recipient_deficit = (
            constraints["local_percentages"][recipient]["min"]
            - composition[recipient]
        )
    else:
        recipient_deficit = 0

    if donor in constraints["local_percentages"]:

        if composition[donor] > constraints["local_percentages"][donor]["max"]:
            donor_surplus = (
                composition[donor]
                - constraints["local_percentages"][donor]["max"]
            )
        elif (
            composition[donor] > constraints["local_percentages"][donor]["min"]
        ):
            donor_surplus = (
                composition[donor]
                - constraints["local_percentages"][donor]["min"]
            )
    else:
        donor_surplus = composition[donor]

    if normal_round(donor_surplus, constraints["digits"] + 1) > 0:
        if recipient_deficit <= 0 or recipient_deficit >= donor_surplus:
            donation = min([0.1, donor_surplus])
        else:
            donation = min([0.1, recipient_deficit])

        # donation = multiple_round(
        #     donation, constraints["percentage_step"], constraints["digits"]
        # )

        # print(
        #     recipient,
        #     "gets",
        #     donation,
        #     "from",
        #     donor,
        # )
        composition.__setitem__(recipient, composition[recipient] + donation)
        composition.__setitem__(donor, composition[donor] - donation)
        # print(composition)
        # print()


def element_is_in_range(
    element: str,
    composition: Alloy.Composition,
    constraints: dict,
    direction: str,
    inclusive: bool = True,
) -> bool:
    """Returns True if an element is within its constrained percentage range,
    False if not.

    :group: alloy.utils

    Parameters
    ----------

    element
        The element to check the constraint range of.
    composition
        The percentages of the alloy.
    constraints
        The constraints on the alloy.
    direction
        Either "min" or "max" depending on which edge of the constraints are
        being checked.
    inclusive
        True if an element should be considered inside the range including
        the endpoint.

    """

    if element not in constraints["local_percentages"]:
        return True

    if element in composition:
        percentage = composition[element]
    else:
        percentage = 0

    if direction == "min":
        if inclusive:
            return (
                percentage
                >= constraints["local_percentages"][element][direction]
            )
        else:
            return (
                percentage
                > constraints["local_percentages"][element][direction]
            )

    elif direction == "max":
        if inclusive:
            return (
                percentage
                <= constraints["local_percentages"][element][direction]
            )

        else:
            return (
                percentage
                < constraints["local_percentages"][element][direction]
            )

    return False


def normal_round(num: float, ndigits: int = 0) -> float:
    """Rounds a float to the specified number of decimal places.

    :group: alloy.utils

    Parameters
    ----------
        num: the value to round
        ndigits: the number of digits to round to
    """
    if ndigits == 0:
        return int(num + 0.5)
    else:
        digit_value = 10**ndigits
        return int(num * digit_value + 0.5) / digit_value


def multiple_round(
    num: float, multiple: Union[float, int], digits=4
) -> Union[float, int]:
    """Round a number to the nearest multiple of another number.

    :group: alloy.utils

    Parameters
    ----------
    num
        Value to round
    multiple
        Multiple-of to round towards
    """
    return round(multiple * round(num / multiple), digits)



"""Entropy related calculations"""

from typing import Union, List
from collections.abc import Iterable
from numbers import Number

import numpy as np

import metallurgy as mg


def ideal_entropy(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the ideal entropy of an alloy. See equation 4 of
    https://doi.org/10.1016/j.chemphys.2020.110898.

    :group: calculations.entropy

    Parameters
    ----------

    alloy
        The alloy for which to calculate the ideal entropy.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [ideal_entropy(a) for a in list(alloy)]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    total_ideal_entropy = 0
    for element in alloy.elements:
        total_ideal_entropy += alloy.composition[element] * np.log(
            alloy.composition[element]
        )

    return -total_ideal_entropy


def ideal_entropy_xia(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns Xia's ideal entropy of an alloy. See equation 8 of
    http://dx.doi.org/10.1063/1.2345259.

    :group: calculations.entropy

    Parameters
    ----------

    alloy
        The alloy for which to calculate Xia's ideal entropy.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [ideal_entropy_xia(a) for a in alloy]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    cube_sum = 0
    for element in alloy.elements:
        volume = mg.periodic_table.elements[element]["atomic_volume"]
        if volume is None:
            return None

        cube_sum += alloy.composition[element] * volume

    ideal_entropy_x = 0
    for element in alloy.composition:
        ideal_entropy_x += alloy.composition[element] * np.log(
            (
                alloy.composition[element]
                * mg.periodic_table.elements[element]["atomic_volume"]
            )
            / cube_sum
        )

    return -ideal_entropy_x


def mismatch_entropy(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the mismatch entropy of an alloy. See equation 2 of
    https://doi.org/10.2320/matertrans1989.41.1372.

    :group: calculations.entropy

    Parameters
    ----------

    alloy
        The alloy for which to calculate the mismatch entropy.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mismatch_entropy(a) for a in alloy]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    if alloy.num_elements == 1:
        return 0.0

    diameters = {}
    for element in alloy.composition:
        if mg.periodic_table.elements[element]["radius"] is None:
            return None

        diameters[element] = mg.periodic_table.elements[element]["radius"] * 2

    sigma_2 = 0
    for element in alloy.composition:
        sigma_2 += alloy.composition[element] * (diameters[element] ** 2)

    sigma_3 = 0
    for element in alloy.composition:
        sigma_3 += alloy.composition[element] * (diameters[element] ** 3)

    y_3 = (sigma_2**3) / (sigma_3**2)

    y_1 = 0
    y_2 = 0

    for i in range(len(alloy.elements) - 1):
        for j in range(i + 1, len(alloy.elements)):
            element = alloy.elements[i]
            other_element = alloy.elements[j]

            y_1 += (
                (diameters[element] + diameters[other_element])
                * ((diameters[element] - diameters[other_element]) ** 2)
                * alloy.composition[element]
                * alloy.composition[other_element]
            )

            y_2 += (
                diameters[element]
                * diameters[other_element]
                * ((diameters[element] - diameters[other_element]) ** 2)
                * alloy.composition[element]
                * alloy.composition[other_element]
            )

    y_1 /= sigma_3

    y_2 *= sigma_2 / (sigma_3**2)

    packing_fraction = 0.64
    zeta = 1.0 / (1 - packing_fraction)

    return (
        ((3.0 / 2.0) * ((zeta**2) - 1) * y_1)
        + ((3.0 / 2.0) * ((zeta - 1) ** 2) * y_2)
        - (1 - y_3) * (0.5 * (zeta - 1) * (zeta - 3) + np.log(zeta))
    )


def mixing_entropy(
    alloy: Union[mg.Alloy, str, dict]
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the mixing entropy of an alloy, combining the
    :func:`~metallurgy.entropy.ideal_entropy` and
    :func:`~metallurgy.entropy.mismatch_entropy`.

    :group: calculations.entropy

    Parameters
    ----------

    alloy
        The alloy for which to calculate the mixing entropy.

    """

    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [mixing_entropy(a) for a in alloy]

    if not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    ideal = ideal_entropy(alloy)
    mismatch = mismatch_entropy(alloy)

    if ideal is not None and mismatch is not None:
        return ideal + mismatch

    return None



from collections import Counter
from math import gcd
from numbers import Number
from string import ascii_uppercase
from typing import List


class Prototype:
    """A crystal structure prototype, defining positions of atoms in a unit cell.

    :group: structures

    Attributes
    ----------

    name
       Name of the structure (e.g. Strukturbericht A1).
    lattice
       Lattice vectors the shape of the unit cell.
    basis
       Basis vectors defining positions of atoms in the unit cell.

    """

    def __init__(
        self,
        name: str,
        lattice: List[List[Number]],
        basis: List[dict],
        space_group: str,
    ):
        self.name = name
        self.lattice = lattice
        self.basis = basis
        self.original_basis = basis
        self.space_group = space_group

    @property
    def num_atoms(self) -> int:
        """Return the number of atoms in the unit cell
        :group: structures
        """
        return len(self.basis)

    @property
    def elements(self) -> list:
        """Return list of unique elements in the unit cell
        :group: structures
        """
        return list(self.element_counts.keys())

    @property
    def element_counts(self) -> dict:
        """Return count of each unique element in the unit cell
        :group: structures
        """
        return Counter(b["element"] for b in self.basis)

    @property
    def composition(self) -> dict:
        """Return percentage of each unique element in the unit cell
        :group: structures
        """
        total = self.num_atoms
        counts = self.element_counts
        return {e: counts[e] / total for e in counts}

    @property
    def num_elements(self) -> int:
        """Return the number of unique elements in the unit cell
        :group: structures
        """
        return len(self.elements)

    @property
    def formula(self) -> str:
        counts = self.element_counts
        divisor = gcd(*counts.values())

        formula_str = ""
        for element, label in zip(counts.keys(), ascii_uppercase):
            formula_str += label
            number = int(counts[element] / divisor)
            if number > 1:
                formula_str += str(number)

        return formula_str
